# coding=utf-8
#
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#               2019-2020 Martin Owens
#               2021 Jonathan Neuhauser, jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Functions for handling styles and embedded css
"""

from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Optional, Generator, TYPE_CHECKING, Tuple
from lxml import etree
import tinycss2
import tinycss2.ast

from .interfaces.IElement import IBaseElement

from .colors import Color
from .properties import (
    _get_tokens_from_value,
    _is_inherit,
    all_properties,
    shorthand_from_value,
    shorthand_properties,
    TokenList,
    _strip_whitespace_nodes,
)
from .css import CSSCompiler, parser

from .utils import FragmentError, NotifyList, NotifyOrderedDict
from .elements._utils import NSS

if TYPE_CHECKING:
    from .elements._base import BaseElement


class Classes(NotifyList):
    """A list of classes applied to an element (used in css and js)"""

    def __init__(self, classes=None, callback=None, element=None):
        if isinstance(classes, str):
            classes = classes.split()
        super().__init__(classes or (), callback=callback)

    def __str__(self):
        return " ".join(self)


@dataclass
class StyleValue:
    """Encapsulates a single parsed style value + its importance state"""

    value: TokenList
    important: bool = False

    def __str__(self):
        return tinycss2.serialize(self.value) + ("!important" if self.important else "")

    def __eq__(self, other):
        return (
            tinycss2.serialize(self.value) == tinycss2.serialize(other.value)
            and self.important == other.important
        )

    def is_inherit(self):
        """Checks if the value is "inherit" """
        return _is_inherit(self.value)


class Style(NotifyOrderedDict):
    """A list of style directives

    .. versionchanged:: 1.2
        The Style API now allows for access to parsed / processed styles via the
        :func:`call` method.

    .. automethod:: __call__
    .. automethod:: __getitem__
    .. automethod:: __setitem__
    """

    color_props = ("stroke", "fill", "stop-color", "flood-color", "lighting-color")
    opacity_props = ("stroke-opacity", "fill-opacity", "opacity", "stop-opacity")
    unit_props = "stroke-width"
    """Dictionary of attributes with units.

    ..versionadded:: 1.2
    """
    associated_props = {
        "fill": "fill-opacity",
        "stroke": "stroke-opacity",
        "stop-color": "stop-opacity",
    }
    """Dictionary of association between color and opacity attributes.

    .. versionadded:: 1.2
    """

    def __init__(self, style=None, callback=None, element=None, **kw):
        self.callback = None
        self.element = element

        # if style is passed as kwargs, replace underscores by dashes
        style = style or [(k.replace("_", "-"), v) for k, v in kw.items()]

        self.update(style)

        # Should accept dict, Style, parsed string, list etc.
        super().__init__(callback=callback)

    def _add(self, key: str, value: StyleValue):
        # Works with both regular dictionaries and Styles
        if key in shorthand_properties:
            chg = shorthand_properties[key].converter.get_shorthand_changes(value.value)  # type: ignore
            for k, v in chg.items():
                self._add(k, StyleValue(v, value.important))
        else:
            if key not in self or (
                not self.get_store(key).important or value.important
            ):
                # Only overwrite if importance of existing value is higher
                super().__setitem__(key, value)

    def _get_val(self, key: str, value):
        if key in all_properties and not isinstance(value, str):
            return StyleValue(
                all_properties[key].converter.convert_back(value, self.element)
            )
        return StyleValue(_get_tokens_from_value(value))

    def _attr_callback(self, key):
        def inner(value):
            self[key] = value

        return inner

    def _parse_str(self, style: str) -> Generator[Tuple[str, StyleValue], None, None]:
        """Create a dictionary from the value of a CSS rule (such as an inline style or
        from an embedded style sheet), including its !important state, in a tokenized
        form. Whitespace tokens from the start and end of the value are stripped.

        Args:
            style: the content of a CSS rule to parse. Can also be a List of
                ComponentValues

        Yields:
            Tuple[str, class:`~inkex.style.StyleValue`]: the parsed attribute
        """
        result = tinycss2.parse_declaration_list(
            style, skip_comments=True, skip_whitespace=True
        )
        for declaration in result:
            if isinstance(declaration, tinycss2.ast.Declaration):
                yield (
                    declaration.name,
                    StyleValue(
                        _strip_whitespace_nodes(declaration.value),
                        declaration.important,
                    ),
                )

    @staticmethod
    def parse_str(style: str, element=None):
        """Parse a style passed as string"""
        return Style(style, element=element)

    def __str__(self):
        """Format an inline style attribute from a dictionary"""
        return self.to_str()

    def to_str(self, sep=";"):
        """Convert to string using a custom delimiter"""
        return sep.join([f"{key}:{value}" for key, value in self.items()])

    def __add__(self, other):
        """Add two styles together to get a third, composing them"""
        ret = self.copy()
        ret.update(Style(other))
        return ret

    def __iadd__(self, other):
        """Add style to this style, the same as ``style.update(dict)``"""
        self.update(other)
        return self

    def __sub__(self, other):
        """Remove keys and return copy"""
        ret = self.copy()
        ret.__isub__(other)
        return ret

    def __isub__(self, other):
        """Remove keys from this style, list of keys or other style dictionary"""
        for key in other:
            self.pop(key, None)
        return self

    def __ne__(self, other):
        return not self.__eq__(other)

    def copy(self):
        """Create a copy of the style.

        .. versionadded:: 1.2"""
        ret = Style({}, element=self.element)
        for key, value in super().items():
            ret[key] = value
        return ret

    def update(self, other):
        """Update, while respecting ``!important`` declarations, and simplifying
        shorthands"""
        if isinstance(other, Style):
            for k, v in super(NotifyOrderedDict, other).items():
                self._add(k, v)
        # Order raw dictionaries so tests can be made reliable
        elif isinstance(other, dict):
            for k, v in sorted(other.items()):
                self._add(k, self._get_val(k, v))

        elif isinstance(other, list) and all(isinstance(i, tuple) for i in other):
            for k, v in other:
                self._add(k, self._get_val(k, v))

        elif isinstance(other, str) or (isinstance(other, list)):
            for k, v in self._parse_str(other):
                self._add(k, v)

    def add_inherited(self, parent):
        """Creates a new Style containing all parent styles with importance "!important"
        and current styles with importance "!important"

        .. versionadded:: 1.2

        Args:
            parent: the parent style that will be merged into this one (will not be
                altered)

        Returns:
            Style: the merged Style object
        """
        ret = self.copy()

        if not (isinstance(parent, Style)):
            return ret

        for item in parent.keys():
            apply = False
            if item in all_properties and all_properties[item].inherited:
                # only set parent value if value is not set or parent importance is
                # higher
                if item not in ret or (
                    not self.get_importance(item) and parent.get_importance(item)
                ):
                    apply = True
            if item in ret and ret.get_store(item).is_inherit():
                apply = True
            if apply:
                super(NotifyOrderedDict, ret).__setitem__(item, parent.get_store(item))
        return ret

    def __setitem__(self, key, value):
        """Sets a style value.

        .. versionchanged:: 1.2
            ``value`` can now also be non-string objects such as a Gradient.

        Args:
            key (str): the attribute name
            value (Any):

                - a :class:`StyleValue`
                - a TokenList (tokenized CSS value),
                - a string with the value
                - any other object. The converter associated with the provided key will
                  attempt to create a string out of the passed value.
        Raises:
            ValueError: when passing something else than string, StyleValue or TokenList
                and key is not a known style attribute
            Error: Other exceptions may be raised when converting non-string objects."""

        if isinstance(value, StyleValue):
            super().__setitem__(key, value)
            return
        if isinstance(value, str):
            value = value.strip()
            tokenized = _get_tokens_from_value(value)

            if key in all_properties:
                all_properties[key].converter.raise_invalid_value(
                    tokenized, self.element
                )
            value = tokenized
        elif (
            isinstance(value, list)
            and len(value) > 0
            and all(isinstance(i, tinycss2.ast.Node) for i in value)
        ):
            pass
        elif key in all_properties:
            value = all_properties[key].converter.convert_back(value, self.element)
        else:
            raise TypeError()
        # Convert value to StyleValue
        super().__setitem__(key, StyleValue(value, False))

    def __getitem__(self, key):
        """Returns the unparsed value of the element (minus a possible ``!important``)

        .. versionchanged:: 1.2
            ``!important`` is removed from the value.
        """
        return tinycss2.serialize(super().__getitem__(key).value)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def get_store(self, key):
        """Gets the :class:`~inkex.properties.BaseStyleValue` of this key, since the
        other interfaces - :func:`__getitem__` and :func:`__call__` - return the
        original and parsed value, respectively.

        .. versionadded:: 1.2

        Args:
            key (str): the attribute name

        Returns:
            BaseStyleValue: the BaseStyleValue struct of this attribute
        """
        return super().__getitem__(key)

    def __call__(self, key: str, element: Optional[BaseElement] = None, default=None):
        """Return the parsed value of a style. Optionally, an element can be passed
        that will be used to find gradient definitions etc.

        .. versionadded:: 1.2"""
        tmp = super().get(key, None)
        v: None | TokenList = None if tmp is None else tmp.value
        if (v is None and (key in all_properties or default is not None)) or (
            v is not None and _is_inherit(v)
        ):  # if the value is still inherit here, return the default
            v = (
                _get_tokens_from_value(default)
                if default is not None
                else (
                    all_properties[key].default_value if key in all_properties else None
                )
            )
            if v is None:
                return v
        if v is not None:
            if key in shorthand_properties:
                return tinycss2.serialize(v)
            if key in all_properties:
                result = all_properties[key].converter.convert(
                    v, element or self.element
                )
            else:
                result = tinycss2.serialize(v)
            if isinstance(result, list) and not isinstance(result, Color):
                result = NotifyList(result, callback=self._attr_callback(key))
            return result
        raise KeyError("Unknown attribute")

    def __eq__(self, other):
        if not isinstance(other, Style):
            other = Style(other)
        if self.keys() != other.keys():
            return False
        for arg in set(self) | set(other):
            if self.get_store(arg) != other.get_store(arg):
                return False
        return True

    def items(self):
        """The styles's items as string

        .. versionadded:: 1.2"""
        for key, value in super().items():
            yield key, tinycss2.serialize(value.value)

    def get_importance(self, key, default=False):
        """Returns whether the declaration with ``key`` is marked as ``!important``

        .. versionadded:: 1.2"""
        if key in self:
            return self.get_store(key).important
        return default

    def set_importance(self, key, importance):
        """Sets the ``!important`` state of a declaration with key ``key``

        .. versionadded:: 1.2"""
        if key in self:
            super().__getitem__(key).important = importance
        else:
            raise KeyError()
        self._callback()

    def get_color(self, name="fill"):
        """Get the color AND opacity as one Color object"""
        color = Color(self.get(name, "none"))
        return color.to_rgba(self.get(name + "-opacity", 1.0))

    def set_color(self, color, name="fill"):
        """Sets the given color AND opacity as rgba to the fill or stroke style
        properties."""
        color = Color(color)
        if color.space == "rgba" and name in Style.associated_props:
            self[Style.associated_props[name]] = color.alpha
            self[name] = color.to_rgb()
        else:
            self[name] = color

    def update_urls(self, old_id, new_id):
        """Find urls in this style and replace them with the new id"""
        for _, elem in super().items():
            for token in elem.value:
                if (
                    isinstance(token, tinycss2.ast.URLToken)
                    and token.value == f"#{old_id}"
                ):
                    token.value = f"#{new_id}"
                    token.representation = f"url(#{new_id})"
                    self._callback()

    def interpolate(self, other, fraction):
        # type: (Style, Style, float) -> Style
        """Interpolate all properties.

        .. versionadded:: 1.1"""
        from .tween import StyleInterpolator
        from inkex.elements import PathElement

        if self.element is None:
            self.element = PathElement(style=str(self))
        if other.element is None:
            other.element = PathElement(style=str(other))
        return StyleInterpolator(self.element, other.element).interpolate(fraction)

    @classmethod
    def cascaded_style(cls, element: BaseElement):
        """Returns the cascaded style of an element (all rules that apply the element
        itself), based on the stylesheets, the presentation attributes and the inline
        style using the respective specificity of the style

        see https://www.w3.org/TR/CSS22/cascade.html#cascading-order

        .. versionadded:: 1.2

        Args:
            element (BaseElement): the element that the cascaded style will be
                computed for

        Returns:
            Style: the cascaded style
        """
        try:
            styles = list(element.root.stylesheets.lookup_specificity(element))
        except FragmentError:
            styles = []

        # presentation attributes have specificity 0,
        # see https://www.w3.org/TR/SVG/styling.html#PresentationAttributes
        styles.append([element.presentation_style(), (0, 0, 0)])

        # would be (1, 0, 0, 0), but then we'd have to extend every entry
        styles.append([element.style, (float("inf"), 0, 0)])

        # sort styles by specificity (ascending, so when overwriting it's correct)
        styles = sorted(styles, key=lambda item: item[1])

        result = styles[0][0].copy()
        for style, _ in styles[1:]:
            result.update(style)
        result.element = element
        return result

    @classmethod
    def specified_style(cls, element):
        """Returns the specified style of an element, i.e. the cascaded style +
        inheritance, see https://www.w3.org/TR/CSS22/cascade.html#specified-value

        .. versionadded:: 1.2

        Args:
            element (BaseElement): the element that the specified style will be computed
                for

        Returns:
            Style: the specified style
        """

        # We currently dont treat the case where parent=absolute value and
        # element=relative value, i.e. specified = relative * absolute.
        cascaded = Style.cascaded_style(element)

        parent = element.getparent()

        if parent is not None and isinstance(parent, IBaseElement):
            cascaded = Style.add_inherited(cascaded, parent.specified_style())
        cascaded.element = element
        return cascaded  # doesn't have a parent

    @classmethod
    def _get_cascade(cls, attribute: str, element: BaseElement) -> Optional[TokenList]:
        if attribute in shorthand_from_value:

            def relevant(style):
                return attribute in style or shorthand_from_value[attribute] in style

        else:

            def relevant(style):
                return attribute in style

        try:
            values = []
            for sheet in element.root.stylesheets:
                for style in sheet:
                    if relevant(style):
                        value = style.get_store(attribute)
                        values += [
                            (value, spec) for spec in style.get_specificities(element)
                        ]
        except FragmentError:
            values = []

        # presentation attributes have specificity 0,
        # see https://www.w3.org/TR/SVG/styling.html#PresentationAttributes
        # they also cannot be shorthands and are always important=False
        if attribute in element.attrib:
            values.append(
                (
                    StyleValue(
                        _get_tokens_from_value(element.attrib[attribute]),
                        False,
                    ),
                    (0, 0, 0),
                )
            )

        if relevant(element.style):
            values.append((element.style.get_store(attribute), (float("inf"), 0, 0)))
        if len(values) == 0:
            return None
        # Sort according to importance, then specificity
        values.sort(key=lambda item: (item[0].important, item[1]))
        return values[-1][0].value

    @classmethod
    def _get_style(cls, attribute: str, element: BaseElement):
        """Specified style for :py:attr:`attribute`"""
        # The resolution order is:
        # - cascade -> then resolve the value, except if the value is "inherit"
        # - parent's computed value
        # - initial (default) value -> then resolve

        result = None
        current = element
        inherited = (
            all_properties[attribute].inherited
            if attribute in all_properties
            else False
        )
        while True:
            result = cls._get_cascade(attribute, current)
            if result is not None and not _is_inherit(result):
                break
            current = current.getparent()
            if current is None or (not inherited and not _is_inherit(result)):
                break
        # Compute value based on current
        if result is None or _is_inherit(result):  # Fallback to default value
            if attribute in all_properties:
                result = all_properties[attribute].default_value
            else:
                return None
        return (
            all_properties[attribute].converter.convert(result, current)
            if attribute in all_properties
            else tinycss2.serialize(result)
        )


class StyleSheets(list):
    """
    Special mechanism which contains all the stylesheets for an svg document
    while also caching lookups for specific elements.

    This caching is needed because data can't be attached to elements as they are
    re-created on the fly by lxml so lookups have to be centralised.
    """

    def lookup(self, element):
        """
        Find all styles for this element.
        """
        for sheet in self:
            for style in sheet.lookup(element):
                yield style

    def lookup_specificity(self, element):
        """
        Find all styles for this element and return the specificity of the match.

        .. versionadded:: 1.2
        """
        for sheet in self:
            for style in sheet.lookup_specificity(element):
                yield style


class StyleSheet(list):
    """
    A style sheet, usually the CDATA contents of a style tag, but also
    a css file used with a css. Will yield multiple Style() classes.
    """

    def __init__(self, content=None, callback=None):
        super().__init__()
        self.callback = None
        # Remove comments
        if content is None:
            parsed = []
        else:
            parsed = tinycss2.parse_stylesheet(
                content, skip_comments=True, skip_whitespace=True
            )
        # Parse rules
        for block in parsed:
            if isinstance(block, tinycss2.ast.QualifiedRule):
                self.append(block)
        self.callback = callback

    def __str__(self):
        return "\n" + "\n".join([str(style) for style in self]) + "\n"

    def _callback(self, style=None):  # pylint: disable=unused-argument
        if self.callback is not None:
            self.callback(self)

    def add(self, rule, style):
        """Append a rule and style combo to this stylesheet"""
        self.append(
            ConditionalStyle(rules=rule, style=str(style), callback=self._callback)
        )

    def append(self, other: str | tinycss2.ast.QualifiedRule):
        """Make sure callback is called when updating"""
        if isinstance(other, str):
            other = tinycss2.parse_one_rule(other)
        if isinstance(other, tinycss2.ast.QualifiedRule):
            other = ConditionalStyle(
                other.prelude, other.content, callback=self._callback
            )
        super().append(other)
        self._callback()

    def lookup(self, element):
        """Lookup the element against all the styles in this sheet"""
        for style in self:
            if any(style.checks(element)):
                yield style

    def lookup_specificity(self, element):
        """Lookup the element_id against all the styles in this sheet
        and return the specificity of the match

        Args:
            element: the element of the element that styles are being queried for

        Yields:
            Tuple[ConditionalStyle, Tuple[int, int, int]]: all matched styles and the
            specificity of the match
        """
        for style in self:
            for specificity in style.get_specificities(element):
                yield (style, specificity)


class ConditionalStyle(Style):
    """
    Just like a Style object, but includes one or more
    conditional rules which places this style in a stylesheet
    rather than being an attribute style.
    """

    def __init__(
        self, rules: str | TokenList = "*", style=None, callback=None, **kwargs
    ):
        super().__init__(style=style, callback=callback, **kwargs)
        self._rules: str | TokenList = rules
        self.rules = list(parser.parse(rules, namespaces=NSS))
        self.checks = [
            CSSCompiler.compile_node(selector.parsed_tree) for selector in self.rules
        ]

    def matches(self, element: etree.Element):
        """Checks if an individual element matches this selector.

        .. versionadded:: 1.4"""
        if isinstance(element, etree._Comment):
            return False
        if any(check(element) for check in self.checks):
            return True
        return False

    def all_matches(self, document: etree.Element):
        """Get all matches of this selector in document as iterator.

        .. versionadded:: 1.4"""
        for el in document.iter():
            if self.matches(el):
                yield el

    def __str__(self):
        """Return this style as a css entry with class"""
        content = self.to_str(";\n  ")
        rules = ",\n".join(str(rule) for rule in self.rules)
        if content:
            return f"{rules} {{\n  {content};\n}}"
        return f"{rules} {{}}"

    def get_specificities(self, element: Optional[BaseElement] = None):
        """Gets an iterator of the specificity of all rules in this ConditionalStyle

        .. versionadded:: 1.2"""
        if element is not None:
            for rule, check in zip(self.rules, self.checks):
                if check(element):
                    yield rule.specificity
        else:
            for rule in self.rules:
                yield rule.specificity



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2020-2022 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Conversion of SVG to XAML"""

import sys
from pathlib import Path

from xml.dom import minidom
from lxml import etree

import inkex
from inkex.localization import inkex_gettext as _


# This bit of import fiddling is taken from
# https://stackoverflow.com/a/28154841/3298143, and it is necessary
# so the script can run both with -m (needed for pypi) as well as
# directly (needed for being an extensions submodule)
if __name__ == "__main__" and __package__ is None:
    file = Path(__file__).resolve()
    parent, top = file.parent, file.parents[1]

    sys.path.append(str(top))
    try:
        sys.path.remove(str(parent))
    except ValueError:  # Already removed
        pass

    __package__ = "inkxaml"

from inkxaml.export.config import GLOBAL_SETTINGS
from inkxaml.xamlobjects import (
    ResourceDictionary,
    DrawingImage,
    KEY_ATTR,
    DrawingBrush,
    Grid,
    BaseElement,
)
from inkxaml.export.structural import SvgElementParser, typemap


def prettify(elem):
    """Return a pretty-printed XML string for the Element."""
    rough_string = etree.tostring(elem, method="xml").decode("utf-8")
    reparsed = minidom.parseString(rough_string)
    return reparsed.toprettyxml(indent=" " * GLOBAL_SETTINGS.indent)


def setup_document():
    """Setup the root element + the root dictionary for swatch resources.
    This setup differs depending on mode and layers-as-resources."""
    # We can reuse the namespace dict, but have to rename the "xaml" item
    nsmap = GLOBAL_SETTINGS.xaml_namespaces.copy()
    nsmap[None] = nsmap.pop("xaml")
    if GLOBAL_SETTINGS.mode == "canvas" and GLOBAL_SETTINGS.layers_as_resources:
        parent = Grid(nsmap=nsmap)
        # We'll use the Grid's ResourceDict for Swatch resources
        rdict = ResourceDictionary(nsmap=nsmap)
        parent.add_property("Resources", rdict)
        GLOBAL_SETTINGS.TOPLEVEL_DICT = rdict
    else:
        # For lowlevel export, this grid will be used directly.
        # For non-layered canvas export, we'll copy the resources out later.
        GLOBAL_SETTINGS.TOPLEVEL_DICT = ResourceDictionary(nsmap=nsmap)
        parent = GLOBAL_SETTINGS.TOPLEVEL_DICT
    return parent


def convert_to_xaml(effect: inkex.OutputExtension, **kwargs):
    """Main conversion method"""
    etree.register_namespace("x", "http://schemas.microsoft.com/winfx/2006/xaml")
    etree.register_namespace("svg", "http://www.w3.org/2000/svg")

    document: inkex.SvgDocumentElement = effect.svg
    GLOBAL_SETTINGS.target = kwargs.get("target", "wpf")
    GLOBAL_SETTINGS.indent = kwargs.get("indent", 4)
    GLOBAL_SETTINGS.mode = kwargs.get("mode", "lowlevel")
    GLOBAL_SETTINGS.swatches = kwargs.get("swatch_treatment", "color")
    GLOBAL_SETTINGS.layers_as_resources = kwargs.get("layers_as_resources", True)
    GLOBAL_SETTINGS.referencing_type = kwargs.get("referencing_type", "DrawingImage")
    xaml_types = {
        "wpf": "http://schemas.microsoft.com/winfx/2006/xaml/presentation",
        "avaloniaui": "https://github.com/avaloniaui",
    }

    GLOBAL_SETTINGS.xaml_namespaces = {
        "x": "http://schemas.microsoft.com/winfx/2006/xaml",
        "d": "http://schemas.microsoft.com/expression/blend/2008",
        "xaml": xaml_types.get(GLOBAL_SETTINGS.target, ""),
        "xml": "http://www.w3.org/XML/1998/namespace",
    }
    GLOBAL_SETTINGS.xaml_namespaces_reversed = dict(
        (b, a) for (a, b) in GLOBAL_SETTINGS.xaml_namespaces.items()
    )

    root_element = setup_document()

    layers = (
        document.xpath("/svg:svg/svg:g[@inkscape:groupmode='layer']")
        if GLOBAL_SETTINGS.layers_as_resources
        else [None]
    )
    for layer in layers:
        layername = layer.get("inkscape:label") if layer is not None else None
        layer_id = layer.get_id() if layer is not None else None
        parser = SvgElementParser(document, layer_id)
        if not GLOBAL_SETTINGS.layers_as_resources:
            # We can fall back to the document name, but we need to strip the .svg
            layername = document.get("sodipodi:docname")
            if layername is not None:
                layername = layername.split(".svg")[0]
        if GLOBAL_SETTINGS.mode == "canvas":
            result = parser.to_canvas_objects()
            if layername is not None:
                result.add_property("Name", BaseElement.sanitize_name(layername))
            if not GLOBAL_SETTINGS.layers_as_resources:
                if len(GLOBAL_SETTINGS.TOPLEVEL_DICT) > 0:
                    result.add_property("Resources", GLOBAL_SETTINGS.TOPLEVEL_DICT)
                root_element = result
            else:
                root_element.add_child(result)
        else:
            result = parser.to_xaml_objects()
            if layername is not None:
                result.add_property(KEY_ATTR, layername)
            root_element.add_child(result)
            key_result = result.get(KEY_ATTR)
            if GLOBAL_SETTINGS.referencing_type == "DrawingImage":
                dwi = DrawingImage()
                dwi.set(KEY_ATTR, "di_" + key_result)
            else:
                dwi = DrawingBrush()
                dwi.set(KEY_ATTR, "db_" + key_result)
                dwi.add_property("Stretch", "Uniform")
            dwi.add_property("Drawing", f"{{StaticResource {key_result}}}")
            root_element.add_child(dwi)
    return prettify(root_element)


class XamlOutput(inkex.OutputExtension):
    """Save as XAML Output"""

    def add_arguments(self, pars):
        pars.add_argument("--mode", choices=["canvas", "lowlevel"], default="lowlevel")
        pars.add_argument("--target", choices=["wpf", "avaloniaui"], default="wpf")
        pars.add_argument(
            "--swatch-treatment",
            choices=["DynamicResource", "StaticResource", "color"],
            default="DynamicResource",
        )
        pars.add_argument("--indent", type=int, default=4)
        pars.add_argument("--layers-as-resources", type=inkex.Boolean, default=False)
        pars.add_argument("--text-to-path", type=inkex.Boolean, default=False)
        pars.add_argument(
            "--referencing-type",
            type=str,
            default="DrawingImage",
            choices=["DrawingImage", "DrawingBrush"],
        )

    def save(self, stream):
        if (
            self.options.mode == "lowlevel"
            or self.options.target == "avaloniaui"
            or (self.options.mode == "canvas" and self.options.text_to_path)
        ):
            # Check if there are any texts that need converting.
            if len(self.svg.xpath("//svg:flowRoot|//svg:text")) > 0:
                self.preprocess(["flowRoot", "text"], unlink_clones=False)

        # Check if the user might accidentally create an empty file.
        warn_toplevel_objects = False
        if self.options.layers_as_resources:
            for elem in self.svg:
                if elem.__class__ in typemap and not isinstance(elem, inkex.Layer):
                    warn_toplevel_objects = True
        if warn_toplevel_objects:
            inkex.errormsg(
                _(
                    'You enabled "Layers as resources", but there are visual '
                    "elements that are not part of a layer and instead contained in the root "
                    "element directly. These elements will be missing in the export. \n\n"
                    "Consider moving them to a layer or disabling this option."
                )
            )
        output = convert_to_xaml(self, **vars(self.options))
        stream.write(output.encode("utf-8"))


def run():
    XamlOutput().run()


if __name__ == "__main__":
    run()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Manpreet Singh <manpreet.singh.dev@proton.me>
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Shared objects for the binary parser.

Shared objects allow accessing fields from a link object with the same id and doc_id
"""

from __future__ import annotations

from dataclasses import InitVar, dataclass, field
from typing import Any, ClassVar, List, Optional
from weakref import WeakValueDictionary

from .types import AFDictObject, AFListObject, AFObject, Document, ObjectStatus


class DocIdInvalidError(Exception):
    def __init__(self, doc_id: int) -> None:
        super().__init__(f"Document id:{doc_id} not found")
        self.doc_id = doc_id


class LinkedObjectNotFound(Exception):
    def __init__(self, obj_id: int) -> None:
        super().__init__(f"Shared object id:{obj_id} not found")
        self.obj_id = obj_id


@dataclass
class SharedAFObject(AFObject):
    _shared_items: ClassVar[List[WeakValueDictionary[int, SharedAFObject]]] = []
    docid: InitVar[Optional[int]] = field(default=None, repr=False)
    _doc_id: int = field(init=False, repr=False, compare=False)

    def __post_init__(self, docid: Optional[int] = None):
        self._doc_id = self.new_doc() if docid is None else docid
        self.set_status(self.status)

    @classmethod
    def new_doc(cls) -> int:
        cls._shared_items.append(WeakValueDictionary())
        return len(cls._shared_items) - 1

    @property
    def shared_items(self) -> WeakValueDictionary[int, SharedAFObject]:
        if self._doc_id >= len(self._shared_items) or self._doc_id < 0:
            raise DocIdInvalidError(self._doc_id)

        return self._shared_items[self._doc_id]

    @property
    def doc_id(self) -> int:
        return self._doc_id

    @property
    def shared_self(self) -> SharedAFObject:
        if self.status != ObjectStatus.LINK:
            raise ValueError("Not a link object")
        if self.id not in self.shared_items:
            raise LinkedObjectNotFound(self.id)
        return self.shared_items[self.id]

    def set_status(self, status: ObjectStatus):
        self.status = status
        if status == ObjectStatus.SHARED:
            if self.id in self.shared_items:
                raise ValueError(f"Duplicate shared object id:{self.id}")
            self.shared_items[self.id] = self
            return

        if self.shared_items.get(self.id) is self:
            self.shared_items.pop(self.id, None)

    def get_type(self) -> Optional[str]:
        return (
            self.shared_self.get_type()
            if self.status == ObjectStatus.LINK
            else super().get_type()
        )

    def __getitem__(self, key: str | int) -> Any:
        if self.status == ObjectStatus.LINK:
            return self.shared_self[key]
        return super().__getitem__(key)  # type: ignore

    def __contains__(self, key: str | int) -> bool:
        if self.status == ObjectStatus.LINK:
            return key in self.shared_self
        return super().__contains__(key)  # type: ignore

    def __len__(self) -> int:
        if self.status == ObjectStatus.LINK:
            return len(self.shared_self)
        return super().__len__()  # type: ignore


@dataclass
class SharedAFDictObject(SharedAFObject, AFDictObject):
    pass


@dataclass
class SharedAFListObject(SharedAFObject, AFListObject):
    pass


@dataclass
class SharedDocument(SharedAFDictObject, Document):
    def set_status(self, status: ObjectStatus):
        assert status == ObjectStatus.ROOT, "Document must be root"
        return super().set_status(status)



#
# Copyright 2018-2022 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""Information about a package"""

from dataclasses import dataclass

from inkex.gui import ChildWindow, TreeView


@dataclass
class ExtensionTreeItem:
    """Shows the name of the item in the extensions tree"""

    name: str
    kind: str


class ExtensionTreeView(TreeView):
    """A list of extensions (inx file based)"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parents = {}
        self.menus = {}

    def setup(self):
        column = self.create_column("Name", expand=True)
        column.add_text_renderer("name")
        self.create_sort(data=lambda item: item.name)
        super().setup()

    def get_menu(self, parent, remain):
        menu_id = "::".join([str(r) for r in remain])
        if menu_id not in self.menus:
            if len(remain) > 1:
                parent = self.get_menu(parent, remain[:-1])

            row = [ExtensionTreeItem(remain[-1], kind="menu")]
            self.menus[menu_id] = self.get_iter(
                self._add_item(row, parent=parent)
            )

        return self.menus[menu_id]

    def add_item(self, item, parent=None):
        if item.kind not in self.parents:
            row = [ExtensionTreeItem(item.kind.title(), kind="category")]
            self.parents[item.kind] = self.get_iter(
                self._add_item(row, parent=None)
            )
        parent = self.parents[item.kind]
        if item.kind == "effect" and len(item.menu) > 1:
            parent = self.get_menu(parent, item.menu[:-1])
        return self._add_item([item], parent)


class MoreInformation(ChildWindow):
    """Show further information for an installed package"""

    name = "info"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.inx = ExtensionTreeView(
            self.widget("inx"), selected=self.select_inx
        )

    def load_widgets(self, pixmaps, item):
        """Initialise the information"""
        try:
            self.widget("info_website").show()
            self.widget("info_website").set_uri(item.link)
            self.widget("info_website").set_label(item.link)
        except Exception:
            self.widget("info_website").hide()

        self.pixmaps = pixmaps
        self.window.set_title(item.name)
        self.widget("info_name").set_label(item.name)
        self.widget("info_desc").set_label(item.summary)
        self.widget("info_version").set_label(item.version)
        self.widget("info_license").set_label(item.license)
        self.widget("info_author").set_label(f"{item.author}")
        try:
            self.widget("info_image").set_from_pixbuf(
                pixmaps.get(item.get_icon())
            )
        except Exception:
            pass

        self.inx.clear()
        self.inx.add(item.get_inx_files())

    def select_inx(self, item):
        pass



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Manpreet Singh <manpreet.singh.dev@proton.me>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import base64
import logging
from typing import List, Optional, Tuple
import math

import inkex
import lxml.etree
from inkex.base import SvgOutputMixin
from inkex import PathElement

from inkaf.parser.extract import AFExtractor
from inkaf.svg import util
from inkaf.svg.enhance import process
from inkaf.svg.fill import (
    AFColor,
    AFGradient,
    AFGradientStop,
    AFGradientType,
    parse_fdsc,
)
from inkaf.svg.raster import AFBitmap, AFImage
from inkaf.svg.util import (
    MAX_CLIP_PATH_RECURSION,
    ClipPathRecursionError,
    get_transformed_bbox,
    interpolate_float,
)
from inkaf.utils import DuplicateFilter, parse

from ..parser.types import AFObject, Document, EnumT
from .curve import AFCurve
from .util import bbox_helper
from .shape import (
    AFCornerPos,
    AFCornerType,
    AFCurveShape,
    AFEllipse,
    AFPie,
    AFPointsShapeBase,
    AFRectangle,
    AFShape,
    AFStar,
    corner_map,
    parse_shape,
)
from .styles import (
    AF_BLEND_MODES,
    SVG_BLEND_MODES,
    AFBlendModeV0,
    AFStroke,
    AFStrokeJoin,
    AFStrokeOrder,
    AFStrokeType,
    parse_fill,
    parse_stroke,
    parse_stroke_color,
    parse_transform,
    stroke_cap_map,
    stroke_join_map,
    stroke_order_map,
)
from .text import (
    TEXT_TYPES,
    AFGlyphAtt,
    AFParaAlign,
    AFParaAtt,
    AFText,
    AFTextLineStyle,
    AFTextType,
)
from .adjustment import adj_dict, Adjustment

logger = logging.getLogger(__name__)
dup_filter = DuplicateFilter()
logger.addFilter(dup_filter)


class AFConverter:
    def __init__(self) -> None:
        self.extractor: AFExtractor
        self.afdoc: Document
        self.doc: lxml.etree._ElementTree
        self.document: inkex.SvgDocumentElement
        self._page_label = ""
        self._has_artboards = False

    def convert(self, extractor: AFExtractor, enhance: bool = True) -> None:
        self.extractor = extractor
        self.afdoc = parse(extractor)
        self.doc = SvgOutputMixin.get_template(
            width=self.afdoc["DocR"].get("DfSz", (100, 100))[0],
            height=self.afdoc["DocR"].get("DfSz", (100, 100))[1],
            unit="px",
        )
        self.document = self.doc.getroot()
        root_chlds = self.afdoc["DocR"].get("Chld", [])
        assert len(root_chlds) <= 1
        if root_chlds:
            self._parse_doc(root_chlds[0])

        if "DCMD" in self.afdoc["DocR"] and "FlNm" in self.afdoc["DocR"]["DCMD"]:
            title = self.afdoc["DocR"]["DCMD"]["FlNm"]
            title = (
                title[: -len(".afdesign")] if title.endswith(".afdesign") else title
            ) + ".svg"
            self.document.add(inkex.Title.new(title))

        if enhance:
            process(self.document)

    def _parse_doc(self, child: AFObject) -> None:
        assert child.get_type() == "Sprd", "Not a document"
        assert not self.document.getchildren()
        self.document.set("width", f"{child['SprB'][2]}")
        self.document.set("height", f"{child['SprB'][3]}")
        self.document.set(
            "viewBox",
            f"{child['SprB'][0]} {child['SprB'][1]} {child['SprB'][2]} {child['SprB'][3]}",
        )

        self._parse_children(child.get("Chld", []))
        self._add_guides(child)

        if self._has_artboards and self._page_label:
            self.document.namedview.get_pages()[0].label = self._page_label

    def _add_guides(
        self, child: AFObject, page_offset: inkex.Vector2d = inkex.Vector2d()
    ) -> None:
        v_guides: list[float] = child.get("GdsV", [])
        h_guides: list[float] = child.get("GdsH", [])

        for pos in v_guides:
            self.document.namedview.add_guide(pos + page_offset.x, orient=False)

        for pos in h_guides:
            self.document.namedview.add_guide(pos + page_offset.y, orient=True)

        # guide color and opacity
        guide_color_af = self.afdoc["DocR"].get("GFil")
        if guide_color_af is None:
            return

        guide_color = parse_fdsc(guide_color_af)
        self.document.namedview.set("guidecolor", self._convert_fill(guide_color))
        self.document.namedview.set(
            "guideopacity",
            (guide_color.alpha if isinstance(guide_color, AFColor) else 1.0),
        )

    def _add_artboard(self, child: AFObject) -> None:
        assert child.get_type() in ["ShpN", "PCrv"], "Invalid artboard type"

        bbox = bbox_helper(child["ShpB" if child.get_type() == "ShpN" else "CvsB"])
        if "Xfrm" in child:
            bbox = get_transformed_bbox(bbox, parse_transform(child["Xfrm"]))

        vx, vy, _, _ = self.document.get_viewbox() or [0, 0, 0, 0]
        self._add_guides(child, inkex.Vector2d(bbox.left - vx, bbox.top - vy))

        if not self._has_artboards:
            self.document.set(
                "viewBox", f"{bbox.left} {bbox.top} {bbox.width} {bbox.height}"
            )
            self.document.set("width", bbox.width)
            self.document.set("height", bbox.height)

            self._page_label = child.get("Desc", "Artboard1")
            self._has_artboards = True
            return

        self.document.namedview.new_page(
            f"{bbox.left - vx}",
            f"{bbox.top - vy}",
            f"{bbox.width}",
            f"{bbox.height}",
            label=child.get(
                "Desc", f"Artboard{len(self.document.namedview._get_pages())}"
            ),
        )

    def _parse_children(
        self,
        children: list[AFObject],
        parent: Optional[inkex.Group] = None,
        parent_af: Optional[AFObject] = None,
        parent_tfm: inkex.Transform = inkex.Transform(),
    ) -> None:
        if parent is None:
            parent = self.document
        assert parent is not None

        for child in children:
            self._add_child(child, parent, parent_af, parent_tfm, False)

    def _add_path_effect(
        self,
        element: inkex.ShapeElement,
        path_effect: inkex.PathEffect,
    ) -> None:
        self.document.defs.add(path_effect)
        path_effect_str = element.get("inkscape:path-effect", "")
        if path_effect_str:
            element.set(
                "inkscape:path-effect", f"{path_effect_str},{path_effect.get_id(1)}"
            )
        else:
            element.set("inkscape:path-effect", f"{path_effect.get_id(1)}")

    def _add_clip(
        self, child: inkex.ShapeElement, parent: inkex.ShapeElement, href: bool = True
    ) -> None:
        assert not isinstance(parent, inkex.elements._groups.GroupBase), (
            f"{parent} of type {type(parent)} is a group"
        )
        assert isinstance(parent, inkex.ShapeElement), (
            f"{parent} of type {type(parent)} is not a ShapeElement"
        )

        if href:
            use = inkex.Use()
            use.href = parent
            use.transform = -child.transform
        else:
            use = parent

        clip = inkex.ClipPath()
        self.document.defs.add(clip)
        clip.add(use)

        child_clip = child
        for i in range(MAX_CLIP_PATH_RECURSION):
            if child_clip.clip is None:
                clip.set("id", f"clip-path{i}-{child.get_id()}")
                child_clip.clip = clip
                return
            assert isinstance(child_clip.clip, inkex.ClipPath)
            child_clip = child_clip.clip

        raise ClipPathRecursionError("Too many nested clip paths")

    def _add_masks(
        self, svg: inkex.ShapeElement, masks: List[AFObject], tfm: inkex.Transform
    ) -> None:
        def fill_pop(element: inkex.ShapeElement) -> None:
            if isinstance(element, inkex.elements._groups.GroupBase):
                for child in element:
                    fill_pop(child)
                return
            element.style.pop("fill")
            element.style.pop("stroke")

        if not masks:
            return

        masks_svg = self.document.defs.add(inkex.Mask(id=f"mask-{svg.get_id()}"))
        curr_mask_svg = masks_svg
        prev_ele: Optional[inkex.ShapeElement] = None
        curr_ele: Optional[inkex.ShapeElement] = None

        for mask_af in masks:
            if mask_af.get_type() in adj_dict:
                try:
                    adj = Adjustment.from_af(mask_af)
                    svg.style("filter").append(adj.to_svg_filter())
                except Exception as e:
                    logger.error(
                        f"Failed to apply adjustment: {e}",
                        exc_info=logger.isEnabledFor(logging.INFO),
                    )

                continue

            if prev_ele is not None:
                curr_mask_svg = self.document.defs.add(
                    inkex.Mask(id=f"mask-{prev_ele.get_id()}")
                )
                prev_ele.set("mask", curr_mask_svg.get_id(2))
            try:
                self._add_child(mask_af, curr_mask_svg, no_catch=True)
            except Exception as e:
                logger.error(
                    f"Failed to parse mask: {e}",
                    exc_info=logger.isEnabledFor(logging.INFO),
                )
                if prev_ele is not None:
                    prev_ele.set("mask", None)
                continue

            curr_ele = list(curr_mask_svg)[-1]
            assert curr_ele is not None
            curr_ele.transform @= tfm
            tfm = tfm @ -curr_ele.transform
            fill_pop(curr_ele)
            curr_ele.style["fill"] = "#ffffff"
            curr_ele.style["stroke"] = "#ffffff"

            prev_ele = curr_ele

        if curr_ele is not None:
            svg.set("mask", masks_svg.get_id(2))

    def _parse_child(self, child: AFObject, tfm: inkex.Transform) -> inkex.ShapeElement:
        # Sprd -> Document
        if child.get_type() == "Sprd":
            raise RuntimeError("Document not at root")

        # ShpN -> Shape
        if child.get_type() == "ShpN":
            scale = util.get_scale_tfm_abs(tfm)
            return self._convert_shape(parse_shape(child["Shpe"], child["ShpB"], scale))

        # PCrv -> Curve
        elif child.get_type() == "PCrv":
            return PathElement.new(
                AFCurve.from_af(child["Crvs"])
                .get_svg_path()
                .transform(tfm, inplace=True)
            )

        # ImgN -> Image
        elif child.get_type() in ("ImgN", "Rstr"):
            img = self._convert_image(AFImage.from_af(child))
            bbox = util.get_transformed_bbox(
                inkex.BoundingBox(
                    (img.left, img.left + img.width), (img.top, img.top + img.height)
                ),
                util.get_scale_tfm_abs(tfm),
            )
            img.set("x", bbox.left)
            img.set("y", bbox.top)
            img.set("width", bbox.width)
            img.set("height", bbox.height)
            return img

        # Scop -> Layer
        elif child.get_type() == "Scop":
            return inkex.Layer()  # TODO: Parse layer properties

        # Grup -> Group
        elif child.get_type() == "Grup":
            return inkex.Group()  # TODO: Parse group properties

        # TxtA -> Text
        elif child.get_type() in TEXT_TYPES:
            return self._convert_text(AFText.from_af(child), tfm)

        else:
            raise NotImplementedError(
                f'Object type "{child.get_type()}" not implemented'
            )

    def _add_child(
        self,
        child: AFObject,
        parent: inkex.Group,
        parent_af: Optional[AFObject] = None,
        parent_tfm: inkex.Transform = inkex.Transform(),
        no_catch: bool = True,
    ) -> None:
        # parent svg object converted from parent_af
        parent_object: Optional[inkex.ShapeElement] = (
            parent
            if parent_af is not None and parent_af.get_type() in ["Scop", "Grup"]
            else parent.getchildren()[0]
            if parent_af is not None and parent.getchildren()
            else None
        )
        tfm = self._combine_tfm(
            self._get_xfrm(child) @ self._get_ftxs(child), parent_tfm
        )
        conversion_failed = True

        style_cascade = [child]
        symbol = False
        try:
            if child.get("ABEn", False):
                self._add_artboard(child)
            try:
                svg = self._parse_child(child, tfm)
            except KeyError:
                # Most likely we have encountered as symbol
                if "SLnk" in child:
                    child = child["SLnk"]["ILOb"][1]
                # Sometimes the symbol wraps the actual object again
                # It appears the properties are grouped
                # into "Master" categories, and we can look them up using them
                # Likely properties end with "Ma" or "M".
                # e.g. StMa, VisM, FiEM, EdMa, HKMs, TMas, TWMa, BGMs, VBFM, VLSM, VTRM
                # These are only relvant when the symbols goes out of sync
                # with the master symbol
                # For most objects the properties are hidden behing BlMM, TxFM is for text
                if "TxFM" in child:
                    child = child["TxFM"]
                elif "BlMM" in child:
                    child = child["BlMM"]
                svg = self._parse_child(child, inkex.Transform())

                style_cascade.insert(0, child)
                symbol = True
        except NotImplementedError as e:
            if no_catch:
                raise e
            logger.error(
                f"Not implemented: {e}", exc_info=logger.isEnabledFor(logging.INFO)
            )
            svg = inkex.Layer.new(label=f"-Failed-{child.get_type()}-")
        except Exception as e:
            if no_catch:
                raise e
            if isinstance(e, KeyError):
                logger.error(
                    f"Expected key missing on {child.get_type()}. Most likely a symbol"
                )
            else:
                logger.exception(f"Failed to convert object type {child.get_type()}")
            svg = inkex.Layer.new(label=f"-Failed-{child.get_type()}-")
        else:
            conversion_failed = False

        if child.get_type() in ["Scop", "Grup"]:
            layer = parent.add(svg)
        else:
            layer = parent.add(inkex.Group())
            layer.add(svg)

        try:
            self._add_common_properties(
                child, svg, parent_tfm, conversion_failed or symbol
            )
            for i, child in enumerate(style_cascade):
                self._add_layer_style(layer, child)
                self._add_style(svg, child, tfm)
                self._add_fx(layer, child)
                self._set_blend_mode(layer, child)
        except NotImplementedError as e:
            logger.error(
                f"Not implemented: {e}", exc_info=logger.isEnabledFor(logging.INFO)
            )
        except Exception:
            logger.exception(f"Failed to set object properties {child.get_type()}")

        if parent_object is not None and not isinstance(parent_object, inkex.Group):
            self._add_clip(layer, parent_object)

        # Vector Masks
        try:
            self._add_masks(layer, child.get("AdCh", []), tfm)
        except Exception:
            logger.exception(f"Failed to add vector masks to #{svg.get_id()}")

        self._parse_children(child.get("Chld", []), layer, child, tfm)

    def _convert_shape(self, child: AFShape) -> inkex.ShapeElement:
        if isinstance(child, AFPointsShapeBase):
            return inkex.Polygon.new(
                points=" ".join([f"{p.x},{p.y}" for p in child.points])
            )
        if isinstance(child, AFRectangle):
            return self._convert_rectangle(child)
        if isinstance(child, AFEllipse):
            return inkex.Ellipse.new(center=child.b_box.center, radius=child.radii)
        if isinstance(child, AFPie):
            if math.isclose(child.start_angle, child.end_angle, abs_tol=1.75e-5):
                child.start_angle = 0
                child.end_angle = math.pi * 2
            outer = PathElement.arc(
                child.b_box.center,
                child.radii.x,
                child.radii.y,
                arctype="slice",
                start=-child.end_angle,
                end=-child.start_angle,
            )
            if not child.inner_radius:
                return outer
            inner = PathElement.arc(
                child.b_box.center,
                child.radii.x * child.inner_radius,
                child.radii.y * child.inner_radius,
                arctype="slice",
                start=-child.end_angle,
                end=-child.start_angle,
            )
            p1 = outer.path.to_absolute()
            p2 = inner.path.reverse().to_absolute()

            if child.start_angle == 0 and child.end_angle == math.pi * 2:
                p = p1 + p2
            else:
                p = (
                    [inkex.paths.Move(*p1[-3].args[-2:])]
                    + [inkex.paths.Line(*p2[2].args)]
                    + p2[3:-1]  # inner arc
                    + [inkex.paths.Line(*p1[0].args)]
                    + p1[1:-2]  # outer arc
                )
            return PathElement.new(p)
        if isinstance(child, AFStar):
            if child.radii.x == child.radii.y and (
                (child.curved_edges and not child.left_curve and not child.right_curve)
                or (
                    not child.curved_edges
                    and not child.inner_circle
                    and not child.outer_circle
                )
            ):
                return PathElement.star(
                    child.b_box.center,
                    (child.radii.x, child.inner_radius_clipped * child.radii.x),
                    child.num_points,
                    args=(
                        -math.pi / 2,
                        -math.pi / 2 + math.pi / child.num_points,
                    ),
                )

        if isinstance(child, AFCurveShape):
            return PathElement.new(child.get_svg_path())

        raise NotImplementedError(f"Conversion for shape {type(child)} not implemented")

    def _convert_rectangle(self, child: AFRectangle) -> inkex.ShapeElement:
        b_box, radii = child.b_box, child.radii

        if any(ct not in corner_map for ct in child.corner_types):
            path_commands = []
            path_commands.append(("M", [b_box.right, b_box.top + radii[1]]))
            path_commands.append(("L", [b_box.right, b_box.bottom - radii[2]]))
            path_commands.extend(child._corner_path(AFCornerPos.BR))
            path_commands.append(("L", [b_box.left + radii[3], b_box.bottom]))
            path_commands.extend(child._corner_path(AFCornerPos.BL))
            path_commands.append(("L", [b_box.left, b_box.top + radii[0]]))
            path_commands.extend(child._corner_path(AFCornerPos.TL))
            path_commands.append(("L", [b_box.right - radii[1], b_box.top]))
            path_commands.extend(child._corner_path(AFCornerPos.TR))
            path_commands.append(("Z", []))
            return PathElement.new(path_commands)

        rect = inkex.Rectangle.new(b_box.left, b_box.top, b_box.width, b_box.height)
        if all(
            r == radii[0] and ct == AFCornerType.ROUNDED
            for r, ct in zip(radii, child.corner_types)
        ) or all(r == 0 for r in radii):
            rect.set("rx", radii[0])
            return rect

        self._add_path_effect(rect, child._corner_path_effect())
        return rect

    def _convert_text(self, text: AFText, tfm: inkex.Transform) -> inkex.TextElement:
        tspans = self._convert_tspans(text, tfm)

        if text.type in [AFTextType.FOLLOW_CURVE, AFTextType.FOLLOW_CURVE2]:
            text_ele = self._convert_text_on_curve(text, tspans)
        else:
            text_ele = self._convert_text_in_shape(text, tspans)

        text_ele.style["white-space"] = "pre"
        text_ele.set("xml:space", "preserve")
        text_ele.style["font-size"] = 0
        text_ele.style["line-height"] = 0
        return text_ele

    def _convert_tspans(self, text: AFText, tfm: inkex.Transform) -> List[inkex.Tspan]:
        para_tspans = []

        for glyph, glyph_atts, para_atts in zip(
            text.glyphs, text.glyph_atts, text.para_atts
        ):
            paras: List[str] = [para + "\n" for para in glyph.split("\u2029")]
            curr_patt_idx = 0
            curr_gatt_idx = 0
            pos_para_offset = 0
            pos_prev_gatt = 0

            for para_str in paras:
                pspan = inkex.Tspan()
                para_tspans.append(pspan)

                max_para_font_size = 0.0
                for glyph_att in glyph_atts.runs[curr_gatt_idx:]:
                    glyph_str = para_str[
                        pos_prev_gatt - pos_para_offset : glyph_att.index
                        - pos_para_offset
                    ]
                    pos_prev_gatt += len(glyph_str)

                    gspan = inkex.Tspan.new(glyph_str.replace("\u0000", ""))
                    pspan.add(gspan)
                    self._add_glyph_style(gspan, glyph_att, tfm)

                    max_para_font_size = max(glyph_att.font_size, max_para_font_size)

                    # End of glyph_att
                    if glyph_att.index <= pos_para_offset + len(para_str):
                        curr_gatt_idx += 1

                    # End of current paragraph (reached \u2029)
                    if glyph_att.index >= pos_para_offset + len(para_str):
                        break

                para_att = para_atts.runs[curr_patt_idx]
                self._add_para_style(
                    pspan,
                    para_att,
                    max_para_font_size,
                )

                # Adjust offests
                pos_para_offset += len(para_str)

                # End of current paragraph_att
                if pos_para_offset >= para_att.index:
                    curr_patt_idx += 1

        return para_tspans

    def _convert_text_on_curve(
        self, text: AFText, tspans: List[inkex.Tspan]
    ) -> inkex.TextElement:
        assert text.text_path is not None
        text_ele = inkex.TextElement.new()
        line = text_ele.add(inkex.TextPath.new(*tspans))
        path = PathElement.new(text.text_path.path.get_svg_path())

        self.document.defs.add(path)
        line.href = path
        line.set("startOffset", f"{text.text_path.starts[0] * 100}%")
        line.set("side", "right" if text.text_path.reverses[0] else "left")
        return text_ele

    def _convert_text_in_shape(
        self,
        text: AFText,
        tspans: List[inkex.Tspan],
    ) -> inkex.TextElement:
        def _convert_shape_inside(
            shape: AFShape | AFCurve, offs: Tuple[float, float], text_type: AFTextType
        ) -> inkex.ShapeElement:
            if text_type != AFTextType.ARTISTIC:
                if isinstance(shape, AFCurve):
                    return PathElement.new(shape.get_svg_path())
                return self._convert_shape(shape)

            assert isinstance(shape, AFRectangle)
            bbox = shape.b_box
            x1, y1, x2, y2 = (
                bbox.left,
                bbox.top,
                bbox.right + offs[0],
                bbox.bottom + offs[1],
            )
            return self._convert_shape(
                AFRectangle(inkex.BoundingBox((x1, x2), (y1, y2)))
            )

        text_ele = inkex.TextElement.new()
        line = text_ele.add(inkex.Tspan.new(*tspans))
        line.set("sodipodi:role", "line")

        bottom_offset = 0
        right_offset = 0

        if tspans:
            top_offset_factor = 0.65
            line1_font_size = max(
                [t.style("font-size") for t in list(tspans[0])] or [0]
            )
            line1_height = (
                float(tspans[0].style("line-height")[:-2]) * top_offset_factor
            )
            tspans[0].style["line-height"] = f"{line1_height}pt"

            bottom_offset = max(line1_height, line1_font_size)
            right_offset = line1_font_size

        shape_inside = _convert_shape_inside(
            text.shape_inside, (right_offset, bottom_offset), text.type
        )
        self.document.defs.add(shape_inside)
        text_ele.style["shape-inside"] = shape_inside
        return text_ele

    def _add_glyph_style(
        self, tspan: inkex.Tspan, glyph_att: AFGlyphAtt, tfm: inkex.Transform
    ) -> None:
        # Fill and stroke
        self._apply_color(tspan, True, glyph_att.fill)
        self._apply_color(tspan, False, glyph_att.stroke_fill)
        self._apply_stroke_style(tspan, glyph_att.stroke, tfm)

        # Font
        tspan.style["font-size"] = glyph_att.font_size
        tspan.style["font-family"] = glyph_att.font.family
        tspan.style["font-weight"] = glyph_att.font.weight
        tspan.style["font-style"] = "italic" if glyph_att.font.italic else "normal"
        tspan.style["font-stretch"] = glyph_att.font.stretch.name.lower()

        if glyph_att.lang:
            tspan.style["lang"] = glyph_att.lang

        if glyph_att.underline != AFTextLineStyle.NONE:
            tspan.style["text-decoration-line"] = "underline"
            tspan.style["text-decoration-style"] = (
                "solid" if glyph_att.underline == AFTextLineStyle.SINGLE else "double"
            )
            if glyph_att.underline_fill:
                tspan.style["text-decoration-fill"] = self._convert_fill(
                    glyph_att.underline_fill
                )

        if glyph_att.line_through != AFTextLineStyle.NONE:
            tspan.style["text-decoration-line"] = "line-through"
            tspan.style["text-decoration-style"] = (
                "solid"
                if glyph_att.line_through == AFTextLineStyle.SINGLE
                else "double"
            )
            if glyph_att.line_through_fill:
                tspan.style["text-decoration-fill"] = self._convert_fill(
                    glyph_att.line_through_fill
                )

    def _add_para_style(
        self, tspan: inkex.Tspan, para_att: AFParaAtt, para_font_size: float
    ) -> None:
        tspan.style["line-height"] = (
            f"{para_att.para_leading.get_line_height(para_font_size)}pt"
        )

        # TODO: Doesn't work
        if para_att.align == AFParaAlign.LEFT:
            tspan.style["text-anchor"] = "start"
            tspan.style["text-align"] = "start"
        elif para_att.align == AFParaAlign.CENTER:
            tspan.style["text-anchor"] = "middle"
            tspan.style["text-align"] = "middle"
        elif para_att.align == AFParaAlign.RIGHT:
            tspan.style["text-anchor"] = "end"
            tspan.style["text-align"] = "end"
        elif para_att.align == AFParaAlign.JUSTIFY_LEFT:
            tspan.style["text-align"] = "justify"
            tspan.style["text-anchor"] = "start"
        elif para_att.align == AFParaAlign.JUSTIFY_CENTER:
            tspan.style["text-align"] = "justify"
            tspan.style["text-anchor"] = "middle"
        elif para_att.align == AFParaAlign.JUSTIFY_RIGHT:
            tspan.style["text-align"] = "justify"
            tspan.style["text-anchor"] = "end"
        elif para_att.align == AFParaAlign.JUSTIFY_ALL:
            tspan.style["text-align"] = "justify"
            tspan.style["text-anchor"] = "middle"
        elif para_att.align in [
            AFParaAlign.TOWARDS_SPLINE,
            AFParaAlign.AWAY_FROM_SPLINE,
        ]:
            raise NotImplementedError(f"Alignment {para_att.align} not implemented")
        else:
            raise NotImplementedError(f"Unknown alignment: {para_att.align}")

    def _convert_image(self, child: AFImage) -> inkex.ShapeElement:
        try:
            if isinstance(child.bitmap.emb_file, str) and child.media_type:
                return self._convert_bitm(child.bitmap, child.media_type)
            elif isinstance(child.bitmap.emb_file, AFObject):
                return self._convert_bitm(child.bitmap)

        except (AssertionError, KeyError, RuntimeError, TypeError):
            logger.warning(
                "Could not load embedded image " + child.bitmap.emb_file
                if isinstance(child.bitmap.emb_file, str)
                else "",
                # exc_info=True,
            )

        path = child.file_path_rel or child.file_path_abs
        if path:
            img = inkex.Image.new(
                x=0, y=0, width=child.bitmap.width, height=child.bitmap.height
            )
            prefix = "file:///" if child.file_path_abs == path else ""
            path = path.replace("\\", "/")
            img.set("xlink:href", f"{prefix}{path}")
            return img

        raise RuntimeError(
            f"Could not load image '{child.bitmap.emb_file if isinstance(child.bitmap.emb_file, str) else ''}'. No relative or absolute path found."
        )

    def _convert_bitm(
        self, child: AFBitmap, mtype: Optional[str] = None
    ) -> inkex.ShapeElement:
        img = inkex.Image.new(x=0, y=0, width=child.width, height=child.height)

        if child.emb_file is not None:
            if isinstance(child.emb_file, str):
                data = self._get_embedded_image_data(child.emb_file)
            elif isinstance(child.emb_file, AFObject):
                mtype = "image/png"
                data = base64.b64encode(
                    AFBitmap.bitmap_from_blocks(child.emb_file, self.extractor)
                ).decode("utf-8")
            img.set(
                "xlink:href",
                f"data:{mtype};base64,{data}",
            )
            return img
        raise NotImplementedError("Bitmap type not supported")

    def _get_embedded_image_data(self, path: str) -> str:
        img_doc = parse(self.extractor, path)
        assert img_doc["Size"] > 0 and len(img_doc["Data"]) == img_doc["Size"], (
            "Invalid embedded image data"
        )
        return base64.b64encode(img_doc["Data"]).decode("utf-8")

    def _add_layer_style(self, svg: inkex.ShapeElement, child: AFObject) -> None:
        # Visibility and Opacity
        # Opac -> layer opacity, FOpc -> filter opacity
        if not child.get("Visi", True):
            svg.style["display"] = "none"
        svg.style["opacity"] = child.get("Opac", 1.0) * child.get("FOpc", 1.0)

    def _add_style(
        self, svg: inkex.ShapeElement, child: AFObject, tfm: inkex.Transform
    ) -> None:
        def adjust_gradient_tfm(grad: AFGradient) -> None:
            """
            Adjusts the gradient transform to account for transformations baked into the
            objects during conversion.
            """
            if child.get_type() in ["ShpN", "ImgN"]:
                grad.tfm = util.get_scale_tfm_abs(tfm) @ grad.tfm
            elif child.get_type() in ["PCrv"]:
                grad.tfm = tfm @ grad.tfm

        fill = parse_fill(child)
        if child.get_type() == "ImgN":
            if fill is not None:
                # TODO: Allow changing image fill
                pass
        else:
            if isinstance(fill, AFGradient):
                adjust_gradient_tfm(fill)
            if fill is not None:
                self._apply_color(svg, True, fill)
            else:
                if "fill" not in svg.style:
                    svg.style["fill"] = None
                    svg.style["fill-opacity"] = 1

        # Stroke
        stroke_fill = parse_stroke_color(child)
        if isinstance(stroke_fill, AFGradient):
            adjust_gradient_tfm(stroke_fill)
        if stroke_fill is not None:
            self._apply_color(svg, False, stroke_fill)

        stroke = parse_stroke(child)
        if stroke is not None:
            strokecopy = None
            if child.get_type() == "ImgN" and stroke.style != AFStrokeType.NONE:
                # Split image and stroke
                strokecopy = inkex.Rectangle.new(
                    svg.get("x"), svg.get("y"), svg.get("width"), svg.get("height")
                )
                strokecopy.transform = svg.transform
                if svg.clip is not None:
                    strokecopy.clip = svg.clip
                strokecopy.style["fill"] = None
                strokecopy.style["stroke"] = svg.style("stroke")
                strokecopy.style["stroke-opacity"] = svg.style("stroke-opacity")
                self._apply_stroke_style(strokecopy, stroke, tfm)
            else:
                self._apply_stroke_style(svg, stroke, tfm)

            power_stroke = stroke.get_power_stroke(
                strokecopy if strokecopy is not None else svg
            )
            if power_stroke is not None:
                if strokecopy is None:
                    if svg.style("fill") is not None:
                        # Split fill and stroke
                        strokecopy = svg.copy()
                    else:
                        strokecopy = svg
                strokecopy.style["fill"] = strokecopy.style["stroke"]
                strokecopy.style["stroke"] = None
                strokecopy.style["fill-opacity"] = strokecopy.style["stroke-opacity"]
                strokecopy.style.pop("stroke-opacity")
                self._add_path_effect(strokecopy, power_stroke)

            if strokecopy is not None:
                svg.style["stroke"] = None
                svg.pop("stroke-opacity")

                if stroke.order == AFStrokeOrder.FRONT:
                    svg.addnext(strokecopy)
                else:
                    svg.addprevious(strokecopy)

        else:
            if "stroke" not in svg.style:
                svg.style["stroke"] = None
                svg.style["stroke-opacity"] = 1

    def _set_blend_mode(self, svg: inkex.ShapeElement, child: AFObject) -> None:
        blend_enum = child.get("Blnd")
        if blend_enum is None:
            return
        assert isinstance(blend_enum, EnumT)

        mode = AF_BLEND_MODES[blend_enum.version](blend_enum.id)
        if mode == AFBlendModeV0.NORMAL:
            return

        if mode.name.lower().replace("_", "-") in SVG_BLEND_MODES:
            svg.style["mix-blend-mode"] = mode.name.lower().replace("_", "-")
            return

        raise NotImplementedError(f"Blend mode {mode.name} not supported")

    def _add_filter(self, svg: inkex.ShapeElement, filt: inkex.Filter) -> None:
        filters = svg.style.get("filter", "")
        svg.style["filter"] = " ".join([filters, f"url(#{filt.get_id()})"])

    def _add_fx(self, svg: inkex.ShapeElement, child: AFObject) -> None:
        for fx in child.get("FiEf", []):
            if not fx.get("Enab", False):
                continue

            filt: inkex.Filter = inkex.Filter()
            self.document.defs.add(filt)
            filt.set("color-interpolation-filters", "sRGB")

            if fx.get_type() == "Gaus":
                radius: float = fx["Radi"] / 3.0
                preserve_alpha: bool = fx["PrAl"]

                filt.add(
                    inkex.Filter.GaussianBlur.new(stdDeviation=radius, result="blur")
                )

                if preserve_alpha:
                    cpt = filt.add(
                        inkex.Filter.ColorMatrix.new(
                            values="1 0 0 0 0 0 1 0 0 0 0 0 1 0 0 0 0 0 100 0",
                            result="blurOpaque",
                        )
                    )
                    cpt.set("in", "blur")

                    composite = filt.add(inkex.Filter.Composite.new(operator="atop"))
                    composite.set("in", "blurOpaque")
                    composite.set("in2", "SourceAlpha")

            else:
                raise NotImplementedError(
                    f"Filter type {fx.get_type()} not implemented"
                )

            self._add_filter(svg, filt)

    def _apply_stroke_style(
        self, svg: inkex.ShapeElement, stroke: AFStroke, tfm: inkex.Transform
    ) -> None:
        # return if stroke type or fill is none
        if stroke.style == AFStrokeType.NONE or svg.style("stroke") is None:
            svg.style["stroke"] = None
            svg.style["stroke-opacity"] = 1
            return

        scale_tfm = util.get_scale_tfm_abs(tfm)
        stroke_width = (
            (stroke.width * (scale_tfm.a + scale_tfm.d) / 2)
            if stroke.scale_with_object
            else stroke.width
        )
        svg.style["stroke-width"] = stroke_width

        if stroke.style == AFStrokeType.DASHED:
            pattern = list(stroke.dash_pattern)
            while pattern[-1] == 0:
                del pattern[-1]
            svg.style["stroke-dasharray"] = [
                ((d if d > 0 else 0.05) * stroke_width) for d in pattern
            ]

        elif stroke.style != AFStrokeType.SOLID:
            raise NotImplementedError(f"Stroke style {stroke.style} not implemented")

        svg.style["stroke-linecap"] = stroke_cap_map[stroke.cap]
        svg.style["paint-order"] = stroke_order_map[stroke.order]

        svg.style["stroke-linejoin"] = stroke_join_map[stroke.join]
        if stroke.join == AFStrokeJoin.MITER:
            svg.style["stroke-miterlimit"] = stroke.miter_clip

    def _apply_color(
        self,
        svg: inkex.ShapeElement,
        fill: bool,
        stroke_fill: Optional[AFColor | AFGradient],
    ):
        prefix = "fill" if fill else "stroke"
        svg.style[prefix] = self._convert_fill(stroke_fill)
        svg.style[f"{prefix}-opacity"] = (
            stroke_fill.alpha if isinstance(stroke_fill, AFColor) else 1.0
        )

    def _convert_fill(self, fill: Optional[AFColor | AFGradient]) -> Optional[str]:
        if fill is None:
            return None
        if isinstance(fill, AFColor):
            return self._convert_color(fill)

        gradient_svg = self._convert_gradient(fill)
        self.document.defs.add(gradient_svg)
        return f"url(#{gradient_svg.get_id()})"

    @staticmethod
    def _convert_gradient(grad: AFGradient) -> inkex.Gradient:
        if grad.grad_type == AFGradientType.LINEAR:
            gradient = inkex.LinearGradient()
            gradient.set("x1", 0)
            gradient.set("y1", 0)
            gradient.set("x2", 1)
            gradient.set("y2", 0)

        elif grad.grad_type in [AFGradientType.RADIAL, AFGradientType.ELLIPTICAL]:
            gradient = inkex.RadialGradient()
            gradient.set("cx", 0)
            gradient.set("cy", 0)
            gradient.set("r", 1)

        else:
            raise NotImplementedError(
                f'Gradient type "{grad.grad_type}" not implemented'
            )

        gradient.set("gradientTransform", grad.tfm)
        gradient.set("gradientUnits", "userSpaceOnUse")

        for stop in AFConverter._convert_stops(grad.stops):
            gradient.add(stop)

        return gradient

    @staticmethod
    def _convert_stops(stops: List[AFGradientStop]) -> List[inkex.Stop]:
        assert len(stops) >= 2, "At least two stops required"
        res = []
        prev_stop = None
        for stop in stops:
            if prev_stop is not None:
                s1, s2 = AFConverter._convert_stop(prev_stop, next_stop=stop)
                res.extend([s1, s2])
            prev_stop = stop
        res.extend(AFConverter._convert_stop(stops[-1]))
        return res

    @staticmethod
    def _convert_stop(
        self, next_stop: AFGradientStop | None = None
    ) -> List[inkex.Stop]:
        def _make_stop(offset: float, color: AFColor) -> inkex.Stop:
            stop = inkex.Stop()
            stop.set("offset", offset)
            stop.style = inkex.Style(
                {
                    "stop-color": AFConverter._convert_color(color),
                    "stop-opacity": color.alpha,
                }
            )
            return stop

        stops = []
        stops.append(_make_stop(self.pos, self.color))
        if next_stop is None:
            return stops

        stops.append(
            _make_stop(
                interpolate_float(self.pos, next_stop.pos, self.mid),
                self.color.interpolate(next_stop.color, 0.5),
            )
        )
        return stops

    @staticmethod
    def _convert_color(color: AFColor) -> str:
        return "#{0:02x}{1:02x}{2:02x}".format(
            *[
                round(c * 255) if isinstance(c, float) else int(c)
                for c in color.to_rgb().color
            ]
        )

    @staticmethod
    def _get_xfrm(child: AFObject) -> inkex.Transform:
        """
        Get the transformation (Xfrm) from the AFObject, if present.

        For text with `shape-inside`, the scale part of Xfrm is applied to the shape.
        The scale for the text itself is present in FTxS.
        """
        return parse_transform(child["Xfrm"]) if "Xfrm" in child else inkex.Transform()

    @staticmethod
    def _get_ftxs(child: AFObject) -> inkex.Transform:
        """
        Get the Frame Text Scale (FTxS) transformation from the AFObject, if present.

        Contains the scale transformation for text objects having `shape-inside`.
        """
        return (
            parse_transform(child["FTxS"])
            if child.get_type() in TEXT_TYPES and "FTxS" in child
            else inkex.Transform()
        )

    @classmethod
    def _add_common_properties(
        cls,
        child: AFObject,
        svg: inkex.ShapeElement,
        parent_tfm: Optional[inkex.Transform] = None,
        conversion_failed: bool = False,
    ) -> None:
        if "Desc" in child and child["Desc"]:
            svg.label = child["Desc"]

        # Xfrm, FTxS -> Transform
        tfm = cls._combine_tfm(cls._get_xfrm(child), parent_tfm or inkex.Transform())

        if conversion_failed and child.get_type() not in ["Scop", "Grup"]:
            svg.transform @= tfm
            return

        # Remove the transformation already baked into the object
        if child.get_type() in TEXT_TYPES - {AFTextType.ARTISTIC.value}:
            shape = svg.style("shape-inside")
            if shape is None:
                shape = svg.getchildren()[0].href
            shape.transform = util.get_scale_tfm_abs(tfm) @ -cls._get_ftxs(child)
            tfm = tfm @ -util.get_scale_tfm_abs(tfm) @ cls._get_ftxs(child)

        elif child.get_type() in ["ShpN", "ImgN"]:
            scale = util.get_scale_tfm_abs(tfm)
            tfm = inkex.Transform(
                (
                    tfm.a / scale.a,
                    tfm.b / scale.a,
                    tfm.c / scale.d,
                    tfm.d / scale.d,
                    tfm.e,
                    tfm.f,
                )
            )

        elif child.get_type() in ["PCrv", "Scop", "Grup"]:
            tfm = inkex.Transform()

        svg.transform @= tfm

    @staticmethod
    def _combine_tfm(
        tfm: inkex.Transform, parent_tfm: inkex.Transform
    ) -> inkex.Transform:
        return parent_tfm @ tfm



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Martin Owens <doctormo@gmail.com>
#                    Sergei Izmailov <sergei.a.izmailov@gmail.com>
#                    Thomas Holder <thomas.holder@schrodinger.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Utilities for parsing SVG documents.

.. versionadded:: 1.2
    Separated out from :py:mod:`inkex.elements._base`
"""

from collections import defaultdict
from typing import DefaultDict, List, Any

from lxml import etree

from ..interfaces.IElement import IBaseElement

from ._utils import splitNS, addNS
from ..utils import errormsg
from ..localization import inkex_gettext as _


class NodeBasedLookup(etree.PythonElementClassLookup):
    """
    We choose what kind of Elements we should return for each element, providing useful
    SVG based API to our extensions system.
    """

    default = IBaseElement

    # (ns,tag) -> list(cls) ; ascending priority
    lookup_table: DefaultDict[str, List[Any]] = defaultdict()

    @classmethod
    def register_class(cls, klass):
        """Register the given class using it's attached tag name"""
        key = addNS(*splitNS(klass.tag_name)[::-1])
        old = cls.lookup_table.get(key, [])
        old.append(klass)
        cls.lookup_table[key] = old

    @classmethod
    def find_class(cls, xpath):
        """Find the class for this type of element defined by an xpath

        .. versionadded:: 1.1
        """
        if isinstance(xpath, type):
            return xpath
        for kls in cls.lookup_table[addNS(*splitNS(xpath.split("/")[-1])[::-1])]:
            # TODO: We could create a apply the xpath attrs to the test element
            # to narrow the search, but this does everything we need right now.
            test_element = kls()
            if kls.is_class_element(test_element):
                return kls
        raise KeyError(f"Could not find svg tag for '{xpath}'")

    def lookup(self, doc, element):  # pylint: disable=unused-argument
        """Lookup called by lxml when assigning elements their object class"""
        try:
            try:
                options = self.lookup_table[element.tag]
            except KeyError:
                if not element.tag.startswith("{"):
                    tag = addNS(*splitNS(element.tag)[::-1])
                    options = self.lookup_table[tag]
                else:
                    return self.default
            for kls in reversed(options):
                if kls.is_class_element(element):  # pylint: disable=protected-access
                    return kls

        except AttributeError:
            # Handle <!-- Comment -->
            return None
        return self.default


SVG_PARSER = etree.XMLParser(huge_tree=True, strip_cdata=False, recover=True)
SVG_PARSER.set_element_class_lookup(NodeBasedLookup())


def load_svg(stream):
    """Load SVG file using the SVG_PARSER"""
    if (isinstance(stream, str) and stream.lstrip().startswith("<")) or (
        isinstance(stream, bytes) and stream.lstrip().startswith(b"<")
    ):
        parsed = etree.ElementTree(etree.fromstring(stream, parser=SVG_PARSER))
    else:
        parsed = etree.parse(stream, parser=SVG_PARSER)
    if len(SVG_PARSER.error_log) > 0:
        errormsg(
            _(
                "A parsing error occurred, which means you are likely working with "
                "a non-conformant SVG file. The following errors were found:\n"
            )
        )
        for __, element in enumerate(SVG_PARSER.error_log):
            errormsg(
                _(
                    "{error_message}. Line {line_number}, column {column_number}",
                ).format(
                    error_message=element.message,
                    line_number=element.line,
                    column_number=element.column,
                )
            )
        errormsg(
            _(
                "\nProcessing will continue; however we encourage you to fix your"
                " file manually."
            )
        )
    return parsed



"""Element abstractions for type comparisons without circular imports

.. versionadded:: 1.2
"""

from __future__ import annotations

from abc import ABC, abstractmethod

import sys
from lxml import etree

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol


class IBaseElement(ABC, etree.ElementBase):
    """Abstraction for BaseElement to avoid circular imports"""

    @abstractmethod
    def get_id(self, as_url=0):
        """Returns the element ID. If not set, generates a unique ID."""
        raise NotImplementedError


class BaseElementProtocol(Protocol):
    """Abstraction for BaseElement, to be used as typehint in mixin classes"""

    def get_id(self, as_url=0) -> str:
        """Returns the element ID. If not set, generates a unique ID."""

    @property
    def root(self) -> ISVGDocumentElement:
        """Returns the element's root."""


class ISVGDocumentElement(IBaseElement):
    """Abstraction for SVGDocumentElement"""



"""
Element based interface provides the bulk of features that allow you to
interact directly with the SVG xml interface.

See the documentation for each of the elements for details on how it works.
"""

from ._utils import addNS, NSS
from ._parser import SVG_PARSER, load_svg
from ._base import ShapeElement, BaseElement
from ._svg import SvgDocumentElement
from ._groups import Group, Layer, Anchor, Marker, ClipPath
from ._polygons import PathElement, Polyline, Polygon, Line, Rectangle, Circle, Ellipse
from ._text import (
    FlowRegion,
    FlowRoot,
    FlowPara,
    FlowDiv,
    FlowSpan,
    TextElement,
    TextPath,
    Tspan,
    SVGfont,
    FontFace,
    Glyph,
    MissingGlyph,
)
from ._use import Symbol, Use
from ._meta import (
    Defs,
    StyleElement,
    Script,
    Desc,
    Title,
    NamedView,
    Guide,
    Metadata,
    ForeignObject,
    Switch,
    Grid,
    Page,
)
from ._filters import (
    Filter,
    Pattern,
    Mask,
    Gradient,
    LinearGradient,
    RadialGradient,
    PathEffect,
    Stop,
    MeshGradient,
    MeshRow,
    MeshPatch,
)
from ._image import Image
